/*
 * This file is a part of KleanSweep.
 *
 * Copyright (C) 2005 Pawel Stolowski <pawel.stolowski@wp.pl>
 *
 * KleanSweep is free software; you can redestribute it and/or modify it
 * under terms of GNU General Public License by Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY. See GPL for more details.
 */

#include "sweeperwizard.h"
#include "sweepscanner.h"
#include <ktabwidget.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qcheckbox.h>
#include <qpushbutton.h>
#include <qvbuttongroup.h>
#include <qhbox.h>
#include <qvbox.h>
#include <qlineedit.h>
#include <qcursor.h>
#include <klocale.h>
#include <kfiledialog.h>
#include <ktar.h>
#include <kmessagebox.h>
#include <kstandarddirs.h>
#include <kurlrequester.h>
#include <unistd.h>
#include <errno.h>
#include <qdatetime.h>
#include <kprocess.h>
#include "logwidget.h"
#include "sweepfileinfo.h"
#include "fileviewitem.h"
#include "filelistview.h" 
#include "filelistviewtarget.h"

SweeperWizard::ScanOpts SweeperWizard::scanopts[] = {
	{SweepScanner::EMPTY_FILE, i18n("Empty files")},
	{SweepScanner::EMPTY_DIR, i18n("Empty directories")},
        {SweepScanner::BROKEN_SYMLINK, i18n("Broken symlinks")},
        {SweepScanner::BACKUP_FILE, i18n("Backup files")},
        {SweepScanner::RPM_ORPHAN, i18n("Orphaned RPM files")},
        {SweepScanner::DEB_ORPHAN, i18n("Orphaned DEB files")},
	{SweepScanner::EMERGE_ORPHAN, i18n("Orphaned Gentoo files")},
        {SweepScanner::DEAD_MENUENTRY, i18n("Dead menu entries")},
	{SweepScanner::DEAD_THUMBNAIL, i18n("Obsolete thumbnails")},
        {SweepScanner::DUPLICATED_FILE, i18n("Duplicated files")},
	{SweepScanner::UNUSED_LOCALE, i18n("Unused locales")},
	{SweepScanner::UNUSED_LOCALE, QString::null}
};

SweeperWizard::SweeperWizard(QWidget *parent): KWizard(parent), scanner(NULL), numchoosen(0)
{
	setCaption("KleanSweep");

	kleansweep_bar1 = new QPixmap(::locate("data", "kleansweep/kleansweep_bar.png"));

        setupWelcomePage();
        setupStartPage();
        setupResultsPage();
	setupArchivePage();
	setupRemovePage();
	setNextEnabled(resultspage, false);
	setNextEnabled(startpage, false);
	setBackEnabled(removepage, false);

	setHelpEnabled(welcomepage, false);
	setHelpEnabled(startpage, false);
	setHelpEnabled(resultspage, false);
	setHelpEnabled(removepage, false);
	setHelpEnabled(archivepage, false);
}

SweeperWizard::~SweeperWizard()
{
	delete scanner;
	delete kleansweep_bar1;
}

void SweeperWizard::setupWelcomePage()
{
        welcomepage = new QWidget(this);

	QHBoxLayout *box0 = new QHBoxLayout(welcomepage, 5, 5);
	addLeftSidePic(welcomepage, box0);

	QString msg = i18n("Welcome to <B>KleanSweep</B>! This application allows you to free some disk space "
				       "by finding unneeded files on your system. You'll be given a chance to backup and "
				       "remove selected files and reclaim your disk space!.<BR><BR><B>Beware! This utility is "
				       "meant for advanced users only. It is assumed that you have some basic understanding "
				       "of what can be safely deleted</B>. The author takes no responsibility "
				       "for any damage caused by using this program. Please report any bugs to pawel.stolowski@wp.pl");
	if (getuid() != 0)
		msg += i18n("<BR><BR><I><B>Note:</B> Since you don't have administrative privileges, you will be "
			    "able to delete only files that you have access to! If you wish to perform system-wide cleaning, "
			    "please run KleanSweep using root account.</I>");
	QLabel *info = new QLabel(msg, welcomepage);
	info->setTextFormat(Qt::RichText);			  
	info->setAlignment(Qt::AlignLeft|Qt::AlignTop|Qt::WordBreak);
	info->setFrameStyle(QFrame::Panel|QFrame::Sunken);
	box0->addWidget(info);

        addPage(welcomepage, "Welcome!");
}

void SweeperWizard::setupStartPage()
{
        startpage = new QWidget(this);
	QHBoxLayout *boxh = new QHBoxLayout(startpage, 5, 5);
	addLeftSidePic(startpage, boxh);
	QVBoxLayout *box0 = new QVBoxLayout(NULL, 0, 5);
	boxh->addLayout(box0);
	
	QLabel *info = new QLabel(i18n("Please select one or more checks to perform. Please keep in mind, than some checks "
	                          "(seeking for duplicated files, RPM and DEB orphaned files) may take a lot of time, "
				  "depending on CPU, memory and disk space. Click <B>Next</B> when ready.") , startpage);
	info->setTextFormat(Qt::RichText);			  
	info->setAlignment(Qt::AlignLeft|Qt::AlignTop|Qt::WordBreak);
	box0->addWidget(info);

	QVButtonGroup *b_group0 = new QVButtonGroup(i18n("Seek for"), startpage);
	box0->addWidget(b_group0);
	for (int i=0; scanopts[i].optname!=QString::null; i++)
	{
		QCheckBox *cb = cb_scan[scanopts[i].ftype] = new QCheckBox(scanopts[i].optname, b_group0);
		connect(cb, SIGNAL(toggled(bool)), this, SLOT(scanCategoryToggled(bool)));
	}
	//
	// disable unfinished checks
	cb_scan[SweepScanner::EMPTY_DIR]->setDisabled(true);
	cb_scan[SweepScanner::UNUSED_LOCALE]->setDisabled(true);
	
	//
	// disable checks for which essential external tools are missing
	if (KStandardDirs::findExe("identify") == QString::null)
		cb_scan[SweepScanner::DEAD_THUMBNAIL]->setDisabled(true);
	if (KStandardDirs::findExe("rpm") == QString::null)
		cb_scan[SweepScanner::RPM_ORPHAN]->setDisabled(true);
	//if (KStandardDirs::findExe("dpkg") == QString::null)
		cb_scan[SweepScanner::DEB_ORPHAN]->setDisabled(true);
	//if (KStandardDirs::findExe("equery") == QString::null)
		cb_scan[SweepScanner::EMERGE_ORPHAN]->setDisabled(true);

	QVButtonGroup *b_group1 = new QVButtonGroup(startpage);
	box0->addWidget(b_group1);
	cb_writable =  new QCheckBox(i18n("Process only files/directories writable for me"), b_group1);
	QHBox *box1 = new QHBox(b_group1);
	box1->setSpacing(5);
	new QLabel(i18n("Start path"), box1);
	url_startpath = new KURLRequester(box1);
	url_startpath->setMode(KFile::Directory | KFile::ExistingOnly | KFile::LocalOnly);
	box0->addStretch();

	//
	// set starting directory
	url_startpath->setURL(getuid() == 0 ? "/" : QDir::homeDirPath());
	
        addPage(startpage, i18n("Select scan options"));
}

void SweeperWizard::setupResultsPage()
{
        resultspage = new KTabWidget(this);
	for (int i=0; scanopts[i].optname!=QString::null; i++)
		views[scanopts[i].ftype] = NULL;
        addPage(resultspage, i18n("Scan results"));
}

void SweeperWizard::setupArchivePage()
{
	archivepage = new QWidget(this);	
	QVBoxLayout *box0 = new QVBoxLayout(archivepage);

	
	QLabel *info = new QLabel(i18n("Selected files will be removed. Please review your choices and click <B>Next</B> when ready. "
				  "It is recommended to <B>create backup archive</B> as there is no other way to restore deleted files."),
			          archivepage);
	info->setTextFormat(Qt::RichText);
	box0->addWidget(info);

	remview = new FileListView(archivepage, NULL);
	box0->addWidget(remview);

	cb_createarch = new QCheckBox(i18n("Create backup archive"), archivepage);
	cb_createarch->setChecked(true);
	box0->addWidget(cb_createarch);

	addPage(archivepage, i18n("Review your choices"));
}

void SweeperWizard::setupRemovePage()
{
	removepage = new QWidget(this);
	QVBoxLayout *box0 = new QVBoxLayout(removepage);
	logwidget = new LogWidget(removepage);
	box0->addWidget(logwidget);
	addPage(removepage, i18n("Sweeping selected files"));
}

void SweeperWizard::addLeftSidePic(QWidget *page, QBoxLayout *parent)
{
	QVBoxLayout *box1 = new QVBoxLayout(parent, 0);

	QLabel *pic = new QLabel(page, 0, Qt::AlignTop);
	pic->setPixmap(*kleansweep_bar1);
	pic->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::MinimumExpanding));
	box1->addWidget(pic, 0, Qt::AlignTop);

	QLabel *fake = new QLabel(page);
	fake->setPaletteBackgroundColor(QColor(0x7d, 0xa0, 0xe7));
	fake->setFixedWidth(kleansweep_bar1->width());
	box1->addWidget(fake);
}

void SweeperWizard::showPage(QWidget *w)
{
        if (w == resultspage)
	{
		const QString path = url_startpath->url();
		if (scanner == NULL)
		{
			scanner = new SweepScanner();
			connect(scanner, SIGNAL(newFileEntry(SweepScanner::FileType, SweepFileInfo *)), this, SLOT(addFileEntry(SweepScanner::FileType, SweepFileInfo *)));
			connect(scanner, SIGNAL(newFileEntry(SweepScanner::FileType, SweepFileInfo *, SweepFileInfo *)), this, SLOT(addFileEntry(SweepScanner::FileType, SweepFileInfo *, SweepFileInfo *)));
			connect(scanner, SIGNAL(finished(bool)), this, SLOT(scannerFinished(bool)));
		}
		FileListView *prev = remview;
		//
		// create or remove views for desired scan types
		for (int i=0; scanopts[i].optname!=QString::null; i++)
		{
			int n;
			if (cb_scan[n = scanopts[i].ftype]->isChecked())
			{
				scanner->enableFileType(scanopts[i].ftype, true);
				if (views[n] == NULL)
				{
					switch (scanopts[i].ftype)
					{
						case SweepScanner::DEAD_THUMBNAIL:
						case SweepScanner::DEAD_MENUENTRY:
						case SweepScanner::BROKEN_SYMLINK: views[n] = new FileListViewTarget(resultspage, prev); break;
						default: views[n] = new FileListView(resultspage, prev); break;
					}
					views[n]->setupColumns();
					connect(views[n], SIGNAL(openDirectory(const QString &)), this, SLOT(openDirectory(const QString &)));
					resultspage->insertTab(views[n], scanopts[i].optname);
				}
				prev = views[n];
			}
			else
			{
				scanner->enableFileType(scanopts[i].ftype, false);
				if (views[n])
				{
					resultspage->removePage(views[n]);
					delete views[n];
					views[n] = NULL;
				}
			}
		}
		//
		// check if types or path were changed since last run of the scanner; if not, do not run the scanner
		if (scanner->settingsChanged() || scanner->lastScanPath() != path) 
		{
			for (int i=0; scanopts[i].optname!=QString::null; i++)
			{
				int n;
				if (views[n = scanopts[i].ftype])
					views[n]->clear(); //clear file views
				remview->clear();
			}
			if (scanner->start(path))
			{
				setNextEnabled(resultspage, false);
				setBackEnabled(resultspage, false);
				setCursor(QCursor(Qt::WaitCursor));
			}
			else
			{
				KMessageBox::sorry(this, i18n("Can't execute helper process. Make sure you have kleansweep-helper in your PATH."));
				delete scanner;
				scanner = NULL;
			}
		}
	}
	else if (w == archivepage)
	{
		remview->clear();
		QPtrList<SweepFileInfo> files = scanner->getFilesForRemoval();
		for (SweepFileInfo *finfo = files.first(); finfo; finfo = files.next())
		{
			new FileViewItem(remview, finfo);
		}
		setNextEnabled(archivepage, files.count() > 0);
	}
	else if (w == removepage)
	{
		QString fname = QDateTime::currentDateTime().toString("yyyyMMdd-hhmmss");
		fname = "kleansweep-" + fname + ".tar.gz";
		bool retry = false;
		do {
			if (cb_createarch->isChecked() || retry)
				fname = KFileDialog::getSaveFileName(fname, "*.tar.gz", NULL, i18n("Select archive"));
			if (fname == QString::null)
			{
				if (KMessageBox::warningYesNo(NULL, i18n("Do you really want to skip archive creation?")) == KMessageBox::Yes)
					break;
				retry = true;
			}
		} while (fname == QString::null);
		bool status = true;
		QPtrList<SweepFileInfo> files = scanner->getFilesForRemoval();
		if (fname != QString::null)
		{
			logwidget->appendOperation(i18n("Creating backup archive ") + " " + fname, LogWidget::STATUS_INFO);

			KTar tar(fname);
			status = tar.open(IO_WriteOnly);
			if (!status)
				logwidget->setOperationStaus(LogWidget::STATUS_FAIL);
			else
			{
				for (SweepFileInfo *finfo = files.first(); finfo; finfo = files.next())
				{
					logwidget->appendOperation(i18n("Archiving") + " " + finfo->url().path());
					status = tar.addLocalFile(finfo->url().path(), finfo->url().path());
					logwidget->setOperationStaus(status ? LogWidget::STATUS_OK : LogWidget::STATUS_FAIL);
					if (!status)
					{
						logwidget->appendOperation(i18n("Aborting files removal due to archiving errors"), LogWidget::STATUS_INFO);
						break;
					}
				}
				tar.close();
			}
		}
		else
		{
			logwidget->appendOperation(i18n("Skipping archive creation"), LogWidget::STATUS_INFO);
		}

		int removed_cnt = 0;
		int notremoved_cnt = 0;
		int removed_size = 0; //total number of bytes reclaimed
		if (status) //continue if no errors occured
		{
			for (SweepFileInfo *finfo = files.first(); finfo; finfo = files.next())
			{
				QString descr(QString::null);
				int size = finfo->size();
				logwidget->appendOperation(i18n("Removing") + " " + finfo->url().path());
				status = (unlink(finfo->url().path().local8Bit()) == 0);
				if (!status)
				{
					switch (errno)
					{
						case EACCES: break; //TODO
						case EPERM: break;
						case ENOENT: break;
						case EIO: break;
						default: break;
					}
				}
				logwidget->setOperationStaus(status ? LogWidget::STATUS_OK : LogWidget::STATUS_FAIL, descr);
				if (status)
				{
					removed_cnt++;
					removed_size += size;
				}
				else
				{
					notremoved_cnt++;
				}
			}
			logwidget->appendOperation(i18n("Successfully removed:") + " " + QString::number(removed_cnt), LogWidget::STATUS_INFO);
			logwidget->appendOperation(i18n("Failed removing:") + " " + QString::number(notremoved_cnt), LogWidget::STATUS_INFO);
			logwidget->appendOperation(i18n("Freed disk space:") + " " + QString::number(removed_size), LogWidget::STATUS_INFO);
		}
		setFinishEnabled(removepage, true);
	}
        KWizard::showPage(w);
}

void SweeperWizard::addFileEntry(SweepScanner::FileType t, SweepFileInfo *f)
{
	if (views[t])
		new FileViewItem(views[t], f);
}

void SweeperWizard::addFileEntry(SweepScanner::FileType t, SweepFileInfo *f, SweepFileInfo *parent)
{
	if (views[t])
	{
		if (FileViewItem *item = views[t]->find(parent))
		{
			new FileViewItem(item, f);
			views[t]->setOpen(item, true);
		}
		else
		{
			new FileViewItem(views[t], f);
		}
	}
}

void SweeperWizard::scannerFinished(bool status)
{
	setNextEnabled(resultspage, true);
	setBackEnabled(resultspage, true);
	setCursor(QCursor());
}

void SweeperWizard::scanCategoryToggled(bool f)
{
	if (f)
		++numchoosen;
	else
		--numchoosen;
	setNextEnabled(startpage, numchoosen>0);
}
		
void SweeperWizard::openDirectory(const QString &path)
{
        KProcess *proc = new KProcess(this);
        *proc << "kfmclient" << "openURL" << path;
        proc->start(KProcess::DontCare, KProcess::NoCommunication);
}

#include "sweeperwizard.moc"
